import moment from 'moment';
import 'tinymce';
import tinymce from 'tinymce';

//declare tinymce = tinymce.EditorManager

export const GUID = {
    
    createNew: function() {
        var uuid;
        if (window.crypto && window.crypto.randomUUID) {
            uuid = window.crypto.randomUUID();
        } else {
            uuid = URL.createObjectURL(new Blob([])).substr(-36);
        }
        var id = BigInt( parseInt( uuid.replaceAll( "-", "" ), 16 ) ).toString( 36 );
        while (id.length<25) {
            id = "0" + id;
        }
        return id;
    }
    
};

export const SHA1 =  {
    
    generateUniqueID: function(inputString) {
        var sha1 = SHA1.generateFromString(inputString);
        var id = BigInt( parseInt( sha1, 16 ) ).toString( 36 );
        while (id.length<25) {
            id = "0" + id;
        }
        return id;
    },
    
    generateFromString: function(msg) {
        
        function rotate_left(n, s) {
            var t4 = (n << s) | (n >>> (32 - s));
            return t4;
        };
        
        function lsb_hex(val) {
            var str = '';
            var i;
            var vh;
            var vl;
            for (i = 0; i <= 6; i += 2) {
                vh = (val >>> (i * 4 + 4)) & 0x0f;
                vl = (val >>> (i * 4)) & 0x0f;
                str += vh.toString(16) + vl.toString(16);
            }
            return str;
        };
        
        function cvt_hex(val) {
            var str = '';
            var i;
            var v;
            for (i = 7; i >= 0; i--) {
                v = (val >>> (i * 4)) & 0x0f;
                str += v.toString(16);
            }
            return str;
        };
        
        function Utf8Encode(string) {
            string = string.replace(/\r\n/g, '\n');
            var utftext = '';
            for (var n = 0; n < string.length; n++) {
                var c = string.charCodeAt(n);
                if (c < 128) {
                    utftext += String.fromCharCode(c);
                } else if ((c > 127) && (c < 2048)) {
                    utftext += String.fromCharCode((c >> 6) | 192);
                    utftext += String.fromCharCode((c & 63) | 128);
                } else {
                    utftext += String.fromCharCode((c >> 12) | 224);
                    utftext += String.fromCharCode(((c >> 6) & 63) | 128);
                    utftext += String.fromCharCode((c & 63) | 128);
                }
            }
            return utftext;
        };
        
        var blockstart;
        var i, j;
        var W = new Array(80);
        var H0 = 0x67452301;
        var H1 = 0xEFCDAB89;
        var H2 = 0x98BADCFE;
        var H3 = 0x10325476;
        var H4 = 0xC3D2E1F0;
        var A, B, C, D, E;
        var temp;
        msg = Utf8Encode(msg);
        
        var msg_len = msg.length;
        var word_array = new Array();
        
        for (i = 0; i < msg_len - 3; i += 4) {
            j = msg.charCodeAt(i) << 24 | msg.charCodeAt(i + 1) << 16 | msg.charCodeAt(i + 2) << 8 | msg.charCodeAt(i + 3);
            word_array.push(j);
        }
        
        switch (msg_len % 4) {
        case 0:
            i = 0x080000000;
            break;
        case 1:
            i = msg.charCodeAt(msg_len - 1) << 24 | 0x0800000;
            break;
        case 2:
            i = msg.charCodeAt(msg_len - 2) << 24 | msg.charCodeAt(msg_len - 1) << 16 | 0x08000;
            break;
        case 3:
            i = msg.charCodeAt(msg_len - 3) << 24 | msg.charCodeAt(msg_len - 2) << 16 | msg.charCodeAt(msg_len - 1) << 8 | 0x80;
            break;
        }
        word_array.push(i);
        while ((word_array.length % 16) != 14) {
            word_array.push(0);
        }
        word_array.push(msg_len >>> 29);
        word_array.push((msg_len << 3) & 0x0ffffffff);
        for (blockstart = 0; blockstart < word_array.length; blockstart += 16) {
            for (i = 0; i < 16; i++) {
                W[i] = word_array[blockstart + i];
            }
            for (i = 16; i <= 79; i++) {
                W[i] = rotate_left(W[i - 3] ^ W[i - 8] ^ W[i - 14] ^ W[i - 16], 1);
            }
            A = H0;
            B = H1;
            C = H2;
            D = H3;
            E = H4;
            for (i = 0; i <= 19; i++) {
                temp = (rotate_left(A, 5) + ((B & C) | (~B & D)) + E + W[i] + 0x5A827999) & 0x0ffffffff;
                E = D;
                D = C;
                C = rotate_left(B, 30);
                B = A;
                A = temp;
            }
            for (i = 20; i <= 39; i++) {
                temp = (rotate_left(A, 5) + (B ^ C ^ D) + E + W[i] + 0x6ED9EBA1) & 0x0ffffffff;
                E = D;
                D = C;
                C = rotate_left(B, 30);
                B = A;
                A = temp;
            }
            for (i = 40; i <= 59; i++) {
                temp = (rotate_left(A, 5) + ((B & C) | (B & D) | (C & D)) + E + W[i] + 0x8F1BBCDC) & 0x0ffffffff;
                E = D;
                D = C;
                C = rotate_left(B, 30);
                B = A;
                A = temp;
            }
            for (i = 60; i <= 79; i++) {
                temp = (rotate_left(A, 5) + (B ^ C ^ D) + E + W[i] + 0xCA62C1D6) & 0x0ffffffff;
                E = D;
                D = C;
                C = rotate_left(B, 30);
                B = A;
                A = temp;
            }
            H0 = (H0 + A) & 0x0ffffffff;
            H1 = (H1 + B) & 0x0ffffffff;
            H2 = (H2 + C) & 0x0ffffffff;
            H3 = (H3 + D) & 0x0ffffffff;
            H4 = (H4 + E) & 0x0ffffffff;
        }
        temp = cvt_hex(H0) + cvt_hex(H1) + cvt_hex(H2) + cvt_hex(H3) + cvt_hex(H4);
        return temp.toLowerCase();
    }
};

export const Util = {

    util: {
        uid: function () {
            return ("0000" + (Math.random() * Math.pow(36, 4) << 0).toString(36)).substr(-4)
        }

    },
    
    normalize: function (name) {
        var normalized = "";
        for (var i = 0; i < name.length; i++) {
            var c = name[i];
            var cc = name.charCodeAt(i);
            if (cc >= 48 && cc <= 57 || cc >= 65 && cc <= 90 || cc >= 97 && cc <= 122) { // 0-9, A-Z, a-z
                normalized += c;
            } else {
                normalized += '_';
                normalized += '' + cc;
            }
        }
        return normalized;
    },
    
    getFilenameFromURL: function(url) {
       if (url) {
           return url.split('/').pop().split('?')[0].split('#')[0];
       }
       return "file";
    },

    getPlainTextForHtmlPanel: function(content) {
        content = Util.getHtmlEncoded(content).replace(/\r/g, "").replace(/\n/g, "<br>");
        content = content.replace(/\t/g, "        ");
        content = content.replace(/\s\s/g, " &nbsp;");
        return content;
    },

    getHtmlBody: function (source) {
        if (source == undefined || source == null) {
            return "";
        }
        var result = source;
        var bodyBegin = result.toLowerCase().indexOf("<body");
        if (bodyBegin != -1) {
            var bodyBeginEnd = result.indexOf(">", bodyBegin);
            var bodyStop = result.toLowerCase().indexOf("</body>", bodyBeginEnd);
            result = result.substring(bodyBeginEnd + 1, bodyStop);
        }
        return result;
    },

    getHtmlEncoded: function (value) {
        if (value == null || value.length == 0) {
            return "";
        }
        return $('<div/>').text(value).html().replace(/{{/g, '((').replace(/}}/g, '))');
    },
    
    getHtmlDecoded: function (value, escape) {
        if (value == undefined || value.length == 0) {
            return "";
        }
        var plain = $('<div/>').html(value).text();
        if (escape) {
            plain = plain.replace(/\&/g, '&amp;');
            plain = plain.replace(/\"/g, '&quot;');
            plain = plain.replace(/\'/g, '&#39;');
        }
        return plain;
    },

    text2html: function (source) {
        return Util.getHtmlEncoded(source).replace(/\n/g, '<br>').replace(/\t/g, '&nbsp;&nbsp;&nbsp;&nbsp;').replace(/\s\s/g, '&nbsp;&nbsp;');
    },
    
    html2plaintext: function(source) {
        var parser = new DOMParser;
        var dom = parser.parseFromString(source, 'text/html');
        return dom.body.textContent;
    },
    
    fastHtml2text: function (source) {
        return $('<div/>').html(source).text();
    },

    html2text: function (source, htmlLight) {
        if (source == undefined || source == null) {
            return "";
        }
        var result = source;

        result = result.replace(/\r/gi, " ");
        result = result.replace(/\n/gi, " ");
        result = result.replace(/\t/gi, "");
        result = result.replace(/\s{2,}/gi, " ");
        result = result.replace(/<\s*head(\s.*?)*>.*?<\/\s*head(\s.*?)*>/gi, "");
        result = result.replace(/<\s*script.*?>.*?<\/\s*script.*?>/gi, "");
        result = result.replace(/<\s*script.*?>/gi, "");
        result = result.replace(/<\s*style.*?>.*?<\/\s*style.*?>/gi, "");
        result = result.replace(/<\s*link.*?>/gi, "");
        result = result.replace(/<\s*html.*?>/gi, "");
        result = result.replace(/<\s*body.*?>/gi, "");
        result = result.replace(/<!--.*?-->/gi, "");

        // Absätze und Umbrüche
        result = result.replace(/<\s*p.*?>/gi, "\n\n");
        result = result.replace(/<\s*br.*?>/gi, "\n");
        result = result.replace(/<\s*div.*?>/gi, "\n");
        result = result.replace(/<\s*h\d.*?>/gi, "\n");
        result = result.replace(/<\s*\/\s*h\d.*?>/gi, "\n");

        //Tabelle
        result = result.replace(/<\s*tr.*?>/gi, "");
        result = result.replace(/<\s*\/\s*tr.*?>/gi, "\n");
        result = result.replace(/<\s*td.*?>/gi, "");
        result = result.replace(/<\s*\/\s*td.*?>/gi, "\t");
        result = result.replace(/<\s*th.*?>/gi, "");
        result = result.replace(/<\s*\/\s*th.*?>/gi, "\t");

        // Listen
        result = result.replace(/<\s*ul.*?>/gi, "\n");
        result = result.replace(/<\s*\/\s*ul.*?>/gi, "\n");
        result = result.replace(/<\s*li.*?>/gi, "* ");
        result = result.replace(/<\s*\/\s*li.*?>/gi, "\n");

        // restliche Tags strippen
        result = result.replace(/<[^>]*>/gi, "");

        // Durch ein <p> oder <br> am Anfang entstehen 1 bis 2 \n, die wir netterweise wegkürzen
        if (htmlLight) {
            result = result.trim();
            
            result = result.replace(/\n\n/gi, "<p class=\"hddefault\">");
            result = result.replace(/\n/gi, "<br>");
            result = result.replace(/\t/gi, "&nbsp;&nbsp;&nbsp;&nbsp;");
        } else {
            if (result.indexOf("\n\n") == 0) {
                result = result.substring(2);
            }
            result = result.replace(/\n\n/gi, "<p class=\"hddefault\">");
            result = result.replace(/<p>/gi, "\n");
        }
        return result;
    },
    parseUrl: function (url) {
        var protEnd = url.indexOf("//");
        url = url.substr(protEnd + 2);
        var domainEnd = url.indexOf("/");
        if (domainEnd >= 0) {
            return url.substr(domainEnd);
        }
        return "";
    },
    getUrlParser : function (path) {
        var parser = document.createElement('a');
        parser.href = path;  
        // parser.protocol; // => "http:"
        // parser.hostname; // => "example.com"
        // parser.port;     // => "3000"
        // parser.pathname; // => "/pathname/"
        // parser.search;   // => "?search=test"
        // parser.hash;     // => "#hash"
        // parser.host;     // => "example.com:3000"
        return parser;
    },
    getSafeHash: function (value) {
        var result = "";
        for (var i = 0; i < value.length; i++) {
            result += "#" + value.charCodeAt(i);
        }
        return result;
    },
    
    /**
     * Escapes html-relevant characters like &, <, >, quotes.
     */
    escapeHtml: function(string) {
    	var entityMap = {
    		    "&": "&amp;",
    		    "<": "&lt;",
    		    ">": "&gt;",
    		    '"': '&quot;',
    		    "'": '&#39;',
    		    "/": '&#x2F;'
    		};
        return String(string).replace(/[&<>"'\/]/g, function (s) {
            return entityMap[s];
        });
    },

    getUniqueFileName: function (fileName, existingNamesArray) {
        var isUnique = false;
        var uniqueName = fileName;
        var index = 0;
        while (!isUnique) {
            isUnique = true;
            angular.forEach(existingNamesArray, function (existingName) {
                if (existingName == uniqueName) {
                    isUnique = false;
                    index++;
                    var pos = fileName.lastIndexOf(".");
                    if (pos == -1) {
                        uniqueName = fileName + "_" + index;
                    } else {
                        uniqueName = fileName.substring(0, pos) + "_" + index + fileName.substring(pos);
                    }
                }
            });
        }
        return uniqueName;
    },
    
    getUniqueName: function (array, element, fieldName) {
        if (fieldName == undefined || fieldName == null) {
            fieldName = "name";
        }
        var isUnique = false;
        var uniqueName = element[fieldName];
        var index = 0;
        while (!isUnique) {
            isUnique = true;
            angular.forEach(array, function (arrayElement) {
                if (arrayElement[fieldName] == uniqueName) {
                    isUnique = false;
                    index++;
                    var pos = element[fieldName].lastIndexOf(".");
                    if (pos == -1) {
                        uniqueName = element[fieldName] + "_" + index;
                    } else {
                        uniqueName = element[fieldName].substring(0, pos) + "_" + index + element[fieldName].substring(pos);
                    }
                }
            });
        }
        return uniqueName;
    },
    
    arrayToString: function (array, suffix, prefix) {
        if (suffix == undefined) {
            suffix = ", ";
        }
        if (prefix == undefined) {
            prefix = "";
        }
        var string = "";
        if (array.length > 0) {
            for (var i = 0; i < array.length; i++) {
                if (i == array.length - 1) {
                    string = string + prefix + array[i];
                    break;
                }
                string = string + prefix + array[i] + suffix;
            }
            ;
        }
        return string;
    },
    
    scrollToCenterOfElement: function (selector, scrollContainerSelector, backupScrollPosition) {
        var element = $(selector);
        var container = $(scrollContainerSelector);
        if (element.offset() == undefined) {
            if (backupScrollPosition != -1) {
                container.scrollTop(backupScrollPosition);
            }
            return;
        }
        var height = element.height();
        var top = parseInt(element.css('top'));
        var scrollTop = container.scrollTop();
        var wHeight = container.height();
        if (top < scrollTop) {
            container.scrollTop(top);
        }
        if (top + height > scrollTop + wHeight) {
            container.scrollTop(top + height - wHeight);
        }
    },
    
    /*
     * http://nealvs.wordpress.com/2010/07/12/how-to-compare-two-javascript-objects-for-equality/
     */
    areEqual: function (x, y) {
        if (x == null) {
            return y == null;
        }
        if (y == null) {
            return x == null;
        }
        // primitives case  (y[p] === y[p]) for x:true and y:false returns true
        if (typeof(x) === typeof(y) && typeof(x) != 'object' && typeof(x) != 'function') {
            return y === x;
        }
        for (var p in y) {
            if (typeof(y[p]) !== typeof(x[p])) return false;
            if ((y[p] === null) !== (x[p] === null)) return false;
            // array case is typeOf object but has no equals method
            if ($.isArray(y[p])) {
                if (!Util.areEqual(y[p], x[p])) {
                    return false;
                }
            }
            switch (typeof(y[p])) {
                case 'undefined':
                    if (typeof(x[p]) != 'undefined') return false;
                    break;
                case 'object':
                    try {
                        if (y[p] !== null && x[p] !== null && (y[p].constructor.toString() !== x[p].constructor.toString() || !y[p].equals(x[p]))) return false;
                    } catch (err) {
                        if (y[p] !== x[p]) return false;
                    }
                    break;
                case 'function':
                    if (p != 'equals' && y[p].toString() != x[p].toString()) return false;
                    break;
                default:
                    if (y[p] !== x[p]) return false;
            }
        }
        return true;
    },

    inArray: function (object, array) {
        for (var i = 0; i < array.length; i++) {
            var hash = array[i].$$hashKey;
            if (hash) {
                object.$$hashKey = hash;
            }
            var equal = Util.areEqual(object, array[i]);
            delete object.$$hashKey;
            if (equal) {
                return equal;
            }
        }
        return false;
    }
}

/**
 * encrypt a json to save transfer to the server
 * @param json the json to encrypt
 * @returns {string} the encrypted json
 */
export function encrypt(json) {
    var l = json.length;
    var result = [];
    var xor = 0x47;
    for (var i = 0; i < l; i++) {
        var code = json.charCodeAt(i);
        var targetCharCode = code ^ xor;
        var targetChar = String.fromCharCode(targetCharCode);
        result.push(targetChar);
        xor = (code & 0xFF);
    }
    return result.join("");
};

/**
 * Returns the json object for the given received data. It will be decrypted and converted if the data is a string. For IE8, the stream is not encrypted, so the object is the resulting instance.
 * @param data the received data from POST request. Needs to be deserialized if it is type of "string"
 * @returns the deserialized json object
 */
export function extractJson(data) {
    if (data && typeof data == "string") {
        var decrypted = decrypt(data);
        try {
            if (decrypted == undefined) {
                return undefined;
            }
            return JSON.parse(decrypted);
        } catch (error) {
            throw error + ": " + Util.html2text(data, false);
        }
    } else {
        return data;
    }
}

export function extractJsonBytes(data) {
    if (data && data instanceof ArrayBuffer) {
        var buffer = new Uint8Array(data);  
        var qwirks = utf8ArrayToStr(buffer);
        if (qwirks.length>0) {
            var decrypted = decrypt(qwirks);
            try {
                if (decrypted == undefined) {
                    return undefined;
                }
                return JSON.parse(decrypted);
            } catch (error) {
                throw error + ": " + Util.html2text(qwirks, false);
            }
        }
    } else {
        return data;
    }
}

export function arrayBufferToBase64( buffer ) {
    var binary = ''
    var bytes = new Uint8Array( buffer )
    var len = bytes.byteLength;
    for (var i = 0; i < len; i++) {
      binary += String.fromCharCode( bytes[ i ] )
    }
    return btoa( binary );
};

export function base64ToArrayBuffer( base64 ) {
    var binary_string = atob( base64 );
    var len = binary_string.length;
    var bytes = new Uint8Array( len );
    for (var i = 0; i < len; i++) {
        bytes[i] = binary_string.charCodeAt(i);
    }
    return bytes.buffer;
}

export function arrayToArrayBuffer( array ) {
    return new Uint8Array(array).buffer;
};

export function arrayToArrayBufferView( array ) {
    return new Uint8Array(array);
};

export function toUTF8Array(str) {
    var utf8 = [];
    for (var i=0; i < str.length; i++) {
        var charcode = str.charCodeAt(i);
        if (charcode < 0x80) utf8.push(charcode);
        else if (charcode < 0x800) {
            utf8.push(0xc0 | (charcode >> 6), 
                      0x80 | (charcode & 0x3f));
        }
        else if (charcode < 0xd800 || charcode >= 0xe000) {
            utf8.push(0xe0 | (charcode >> 12), 
                      0x80 | ((charcode>>6) & 0x3f), 
                      0x80 | (charcode & 0x3f));
        }
        // surrogate pair
        else {
            i++;
            // UTF-16 encodes 0x10000-0x10FFFF by
            // subtracting 0x10000 and splitting the
            // 20 bits of 0x0-0xFFFFF into two halves
            charcode = 0x10000 + (((charcode & 0x3ff)<<10)
                      | (str.charCodeAt(i) & 0x3ff))
            utf8.push(0xf0 | (charcode >>18), 
                      0x80 | ((charcode>>12) & 0x3f), 
                      0x80 | ((charcode>>6) & 0x3f), 
                      0x80 | (charcode & 0x3f));
        }
    }
    return utf8;
};

export const utf8ArrayToStr = (function () {
    var charCache = new Array(128);  // Preallocate the cache for the common single byte chars
    var charFromCodePt = String.fromCodePoint || String.fromCharCode;
    var result = [];

    return function (array) {
        var codePt, byte1;
        var buffLen = array.length;

        result.length = 0;

        for (var i = 0; i < buffLen;) {
            byte1 = array[i++];

            if (byte1 <= 0x7F) {
                codePt = byte1;
            } else if (byte1 <= 0xDF) {
                codePt = ((byte1 & 0x1F) << 6) | (array[i++] & 0x3F);
            } else if (byte1 <= 0xEF) {
                codePt = ((byte1 & 0x0F) << 12) | ((array[i++] & 0x3F) << 6) | (array[i++] & 0x3F);
            } else {
                codePt = ((byte1 & 0x07) << 18) | ((array[i++] & 0x3F) << 12) | ((array[i++] & 0x3F) << 6) | (array[i++] & 0x3F);
                if (!String.fromCodePoint) {
                    //Converting to UTF-16
                    codePt -= 0x10000;
                    var surrogate = 0xD800 | (codePt >> 10 ); // high Surrogate
                    result.push(charCache[surrogate] || (charCache[surrogate] = charFromCodePt(surrogate)));
                    codePt = 0xDC00 | ( codePt & 0x3FF ); // low Surrogate
                }
            }

            result.push(charCache[codePt] || (charCache[codePt] = charFromCodePt(codePt)));
        }

        return result.join('');
    };
})();

/**
 * decrypt a json from the server
 * @param qwirks the string to decrypt
 * @returns {string} the decrypted json
 */
export function decrypt(qwirks) {
    if (qwirks && qwirks.charCodeAt(0) === 123 /* '{' */) {
        return qwirks;
    }
    if (qwirks && qwirks == "null") {
        return undefined;
    }
    var l = qwirks.length;
    var result = [];
    var xor = 0x47;
    for (var i = 0; i < l; i++) {
        var code = qwirks.charCodeAt(i);
        xor = (code ^ xor);
        result.push(String.fromCharCode(xor));
        xor &= 0xFF;
    }
    return result.join("");
};

export const ColorUtil = {
    /* accepts parameters
     * h  Object = {h:x, s:y, v:z}
     * OR
     * h, s, v
     *
     * 0 <= h, s, v <= 1
     */
    HSVtoRGB: function (h, s, v) {
        var r, g, b, i, f, p, q, t;
        if (h && s === undefined && v === undefined) {
            s = h.s, v = h.v, h = h.h;
        }
        i = Math.floor(h * 6);
        f = h * 6 - i;
        p = v * (1 - s);
        q = v * (1 - f * s);
        t = v * (1 - (1 - f) * s);
        switch (i % 6) {
            case 0:
                r = v, g = t, b = p;
                break;
            case 1:
                r = q, g = v, b = p;
                break;
            case 2:
                r = p, g = v, b = t;
                break;
            case 3:
                r = p, g = q, b = v;
                break;
            case 4:
                r = t, g = p, b = v;
                break;
            case 5:
                r = v, g = p, b = q;
                break;
        }
        return {
            r: Math.floor(r * 255),
            g: Math.floor(g * 255),
            b: Math.floor(b * 255)
        };
    },

    componentToHex: function (c) {
        var hex = c.toString(16);
        return hex.length == 1 ? "0" + hex : hex;
    },

    rgbToHex: function (r, g, b) {
        return "#" + ColorUtil.componentToHex(r) + ColorUtil.componentToHex(g) + ColorUtil.componentToHex(b);
    },
    
    parseRGB : function(rgb) {
	  rgb = rgb.match(/^rgba?\((\d+),\s*(\d+),\s*(\d+)(?:,\s*([\d\.]+))?\)$/);

	  function hex(x) { return ("0" + parseInt(x).toString(16)).slice(-2); };
	  
	  return (hex(rgb[1]) + hex(rgb[2]) + hex(rgb[3])).toUpperCase();
	},
    
    hexToRgba: function(hex, alpha){
	    var c;
	    if(/^#([A-Fa-f0-9]{3}){1,2}$/.test(hex)){
	        c= hex.substring(1).split('');
	        if(c.length== 3){
	            c= [c[0], c[0], c[1], c[1], c[2], c[2]];
	        }
	        c= '0x'+c.join('');
	        return 'rgba('+[(c>>16)&255, (c>>8)&255, c&255].join(',')+','+alpha+')';
	    }
	    throw new Error('Bad Hex');
	},
    
    mix : function(color_1, color_2, weight) {
	  function d2h(d) { return d.toString(16); }  // convert a decimal value to hex
	  function h2d(h) { return parseInt(h, 16); } // convert a hex value to decimal 

	  weight = (typeof(weight) !== 'undefined') ? weight : 50; // set the weight to 50%, if that argument is omitted

	  var color = "#";

	  for(var i = 0; i <= 5; i += 2) { // loop through each of the 3 hex pairs—red, green, and blue
	    var v1 = h2d(color_1.substr(i, 2)), // extract the current pairs
	        v2 = h2d(color_2.substr(i, 2)),
	        
	        // combine the current pairs from each source color, according to the specified weight
	        val = d2h(Math.floor(v2 + (v1 - v2) * (weight / 100.0))); 

	    while(val.length < 2) { val = '0' + val; } // prepend a '0' if val results in a single digit
	    
	    color += val; // concatenate val to our new color string
	  }
	    
	  return color; // PROFIT!
	},

    getPriorityColor: function (priorityIndex, priorityCount) {
        var maxHue = 100;
        var minHue = 0;

        var hue = maxHue;
        var saturation = 1;
        var value = 1;
        if (priorityCount > 0) {
            hue = maxHue - ((maxHue - minHue) / (priorityCount - 1) * priorityIndex);
        }
        var color = ColorUtil.HSVtoRGB(hue / 360, saturation, value);
        return ColorUtil.rgbToHex(color.r, color.g, color.b);
    },
	
	getColorForID: function (id) {
        var string = id;
        
        var normalizeChar = function(c) {
            if( c >= 48 && c <= 57 ) {
                c -= 48;
            }
            if( c >= 97 && c <= 122 ) {
                c -= 87;
            }
            return c;
        };

        var h = normalizeChar( string.codePointAt(1) ) / 35.0;
        var s = normalizeChar( string.codePointAt(9) ) / 35.0;
        var b = normalizeChar( string.codePointAt(17) ) / 35.0;

        // check whether starting with 9 digits, used by helpdesk for guid generation on database
        var startWithLegacyDigits = true;
        for( var i = 0; i < 9; i++ ) {
            var c = string.codePointAt(i);
            if( c < 48 || c > 57 ) {
                startWithLegacyDigits = false;
                break;
            }
        }

        if( startWithLegacyDigits ) {

            // check whether only hex characters are included
            var onlyHexCharacters = true;
            for( var i = 9; i < 25; i++ ) {
                var c = string.codePointAt(i);
                if( c > 102 ) { // > 'f'
                    onlyHexCharacters = false;
                    break;
                }
            }

            if( onlyHexCharacters ) {
                // for 0-9 a-f we use touples and scale their value to 0..35
                var factor = 36 / 165.0;
                h = Math.round( ((normalizeChar( string.codePointAt(12) ) * 10) + normalizeChar( string.codePointAt(13) )) * factor ) / 35.0;
                s = Math.round( ((normalizeChar( string.codePointAt(17) ) * 10) + normalizeChar( string.codePointAt(18) )) * factor ) / 35.0;
                b = Math.round( ((normalizeChar( string.codePointAt(23) ) * 10) + normalizeChar( string.codePointAt(24) )) * factor ) / 35.0;
            } else {
                // we need to take different positions for the characters, ignoring the first 9 characters that are only a representation of the int id in helpdesk
                h = normalizeChar( string.codePointAt(10) ) / 35.0;
                s = normalizeChar( string.codePointAt(17) ) / 35.0;
                b = normalizeChar( string.codePointAt(24) ) / 35.0;
            }
        }

        h *= 0.98;
        h += 0.01;
        s *= 0.6;
        s += 0.3;
        b /= 2.0;
        b *= s;
        b += 0.5;
        
        return ColorUtil.HSVtoRGB(h,s,b);
    }

}

export const FieldUtil = {

    getDisplayValue: function (ticketModel, data, field) {
        if (data == undefined) {
            return "";
        }
        var value = data[field];
        if (value == undefined || (value == "" && field != "TICKETDATA_ITILID" && field != "TICKETDATA_STATUSID" )) {
            return "";
        }
        switch (field) {
            case 'TICKETDATA_TICKETID' :
                return '#' + value;
            case 'TICKETDATA_LASTMODIFIED' :
                return ValueFormatter.formatDate(new Date(parseInt(value)), undefined);
            case 'TICKETDATA_INQUIRYDATE' :
                return ValueFormatter.formatDate(new Date(parseInt(value)), undefined);
            case 'TICKETDATA_DEADLINE' :
                return ValueFormatter.formatDate(new Date(parseInt(value)), undefined);
            case 'TICKETDATA_DUETIME' :
                return ValueFormatter.formatTime(parseInt(value));
            case 'TICKETDATA_SUMTIMES' :
                return ValueFormatter.formatTime(parseInt(value));
        }

        var display = undefined;
        if (data[field + '_DISPLAYNAME'] != undefined) {
            display = data[field + '_DISPLAYNAME'];
        } else if (field == 'TICKETDATA_PRIORITYID') {
            var prio = ticketModel.getPriority(data[field]);
            if (prio != undefined) {
                display = prio.displayName;
            }
        } else if (field == 'TICKETDATA_STATUSID') {
            var state = ticketModel.getState(data[field]);
            if (state != undefined) {
                display = state.displayName;
            }
        } else if (field == 'TICKETDATA_ITILID') {
            var itil = ticketModel.getItil(data[field]);
            if (itil != undefined) {
                display = itil.displayName;
            }
        } else if (field == 'TICKETDATA_CLASSIFICATIONID') {
            var classification = ticketModel.getClassification(data[field]);
            if (classification != undefined) {
                display = classification.displayName;
            }
        }

        if (display == undefined) {
            display = data[field];
        }
        return display;
    },

    getDataID: function (field) {
        switch (field) {
            case 'TICKETDATA_STATUSID':
                return 0;
            case 'TICKETDATA_RESOURCEID':
                return 1;
            case 'TICKETDATA_PRIORITYID':
                return 2;
            case 'TICKETDATA_CLASSIFICATIONID':
                return 3;
            case 'TICKETDATA_CATEGORY':
                return 4;
            case 'USERDATA_LOCATION':
                return 5;
            case 'USERDATA_GROUP':
                return 6;
            case 'TICKETDATA_ITILID':
                return 7;
            // case '': return 8;   // TYPE_FREI = 8; k.A. wie man das rückwärts auflösen loss
            case 'TICKETDATA_LASTTEXTEDITOR':
                return 9;
            case 'TICKETDATA_TICKETFIELD1':
                return 10;
            case 'TICKETDATA_TICKETFIELD2':
                return 11;
            case 'TICKETDATA_TICKETFIELD3':
                return 12;
        }
        return -1;
    },

    convertAttachmentDescription: function (att) {
        var extension = att.fileExtension;
        var mimExt = extension;
        if (mimExt.indexOf('.') == 0) {
            mimExt = mimExt.substr(1);
        }
        return {
            name: att.fileDisplayName,
            lastModified: ValueFormatter.formatDate(new Date(att.lastModified), undefined),
            lastModifiedValue: att.lastModified,
            attachmentType: 'Attachment',
            size: ValueFormatter.formatFileSize(att.filesize),
            sizeValue: att.filesize,
            id: att.ticketId + '_' + att.stepId + '_' + Util.util.uid(),
            idValue: att.ticketId,
            mimetype: FieldUtil.isImageFile(extension) ? "image/" + mimExt : (FieldUtil.isTxtFile(extension) ? 'text/' + mimExt : 'application/x-unknown'),
            serverpath: att.serverFilePath,
            stepId: att.stepId,
            type: att.attachmentType
        };
    },

    isImageFile: function (filename) {
        filename = filename.toLowerCase();
        var index = filename.lastIndexOf('.');
        if (index >= 0 && index < filename.length - 1) {
            filename = filename.substring(index + 1, filename.length);
        }
        return filename == "jpg" || filename == "jpeg" || filename == "gif" || filename == "bmp" || filename == "png";
    },
    
    isImageFileType: function (type) {
        type = type.toLowerCase();
        if(type.indexOf("image/") == 0) {
            type = type.substr(6);
            return type == "jpeg" || type == "gif" || type == "bmp" || type == "tiff" || type == "png"
            
        }
        return false;
    },

    isTxtFile: function (filename) {
        filename = filename.toLowerCase();
        var index = filename.lastIndexOf('.');
        if (index >= 0 && index < filename.length - 1) {
            filename = filename.substring(index + 1, filename.length);
        }
        return filename == "txt" || filename == "log" || filename == "htm" || filename == "html" || filename == "xml" || filename == "js" || filename == "msg" || filename == "eml" || filename == "ini" || filename == "java" || filename == "css" || filename == "less" || filename == "json";
    },

    removeObsoleteEmbeddedImages: function (htmlTextArray, attachments) {
        var returnArray = true;
        if (!$.isArray(htmlTextArray)) {
            htmlTextArray = [htmlTextArray];
            returnArray = false;
        }
        for (var htmlIndex=0; htmlIndex<htmlTextArray.length; htmlIndex++) {
            var htmlText = htmlTextArray[htmlIndex];
            if (htmlText == undefined) {
                continue;
            }
            htmlText = htmlText.replace(/data-mce.*?=".*?"/gi, "");
            var matched = htmlText.match(/<img id="(embedded_.*?)".*?>/gi);
            if (matched != null && matched != undefined) {
                for (var i = 0; i < matched.length; i++) {
                    var mat = matched[i];
                    var newMatch = mat.replace(/ src=".*?"/gi, " src=\"\"");
                    htmlText = htmlText.replace(mat, newMatch);
                }
            }
            htmlTextArray[htmlIndex] = htmlText;
        }
        for (var i = attachments.length - 1; i >= 0; i--) {
            var att = attachments[i];
            if (att.attachmentType == "EmbeddedImage") {
                var attFound = false;
                if (att.id.indexOf('embedded_') != 0) {
                    attFound = true;
                } else {
                    for (var htmlIndex=0; htmlIndex<htmlTextArray.length; htmlIndex++) {
                        htmlText = htmlTextArray[htmlIndex];
                        if (htmlText == undefined) {
                            continue;
                        }
                        if (htmlText.indexOf("<img id=\"" + att.id + "\"") >= 0) {
                            var reg = new RegExp(" id=\"" + att.id + "\"", "gi");
                            htmlText = htmlText.replace(reg, " id=\"" + att.name + "\" src=\"" + att.name + "\"");
                            htmlText = htmlText.replace(/ src=\"\"/gi, "");
                            attFound = true;
                            htmlTextArray[htmlIndex] = htmlText;
                        }
                    }
            }
                if (!attFound) {
                    attachments.splice(i, 1);
                }
            }
        }
        return returnArray ? htmlTextArray : htmlTextArray[0];
    }

}

export const JsonUtil = {

    checkForDifferences: function (data, refData) {
        return angular.toJson(data) != angular.toJson(refData);   
    },

    createSnapshot: function (data) {
        var retValue = jQuery.extend(true, {}, data);
        if (!JsonUtil.checkForDifferences(data, retValue)) {
            return retValue;
        }
        return undefined;
    }
};

export const FormFieldUtil = {

    checkForDifferences: function (model, initialData) {
        var currentData = FormFieldUtil.createSnapshot(model);

        var diff = {};
        for (var i = 0, j = currentData.length; i < j; i++) {
            var cd = currentData[i];
            var id = initialData[i];
            if (cd.value != id.value) {
                diff[cd.key] = cd.value;
            }
        }

        return diff;
    },

    createSnapshot: function (model) {
        var result = [];
        var keys = Object.keys(model);
        for (var g = 0, h = keys.length; g < h; g++) {
            var propertyGroupKey = keys[g];
            var propertyGroup = model[propertyGroupKey];
            for (var i = 0, j = propertyGroup.length; i < j; i++) {
                var m = propertyGroup[i];
                var value = FormFieldUtil.getStringRepresentation(m.value, false);
                result.push({key: m.key, value: value});
            }
        }
        return result;
    },
    
    checkForHDDifferences: function (model, initialData) {
        if (typeof tinymce != 'undefined' && tinymce.activeEditor != undefined) {
            tinymce.activeEditor.save()
        }

        var diff = [];
        if (model == undefined) {
            return diff;
        }
        
        var currentData = FormFieldUtil.createHDSnapshot(model);
        for (var i = 0, j = currentData.length; i < j; i++) {
            var cd = currentData[i];
            var id = initialData[i];
            if (id == undefined || cd.value != id.value) {
                diff.push(cd);
            }
        }

        return diff;
    },
    
    createHDSnapshot: function (model) {
        var result = [];
        for (var i = 0, j = model.length; i < j; i++) {
            var m = model[i];
            result.push({field: m.field, value: String(m.value)});
        }
        return result;
    },

    getStringRepresentation: function (value, asJson) {
        if (value == null || value == undefined) {
            return "null";
        }
        if ($.isArray(value)) {
            return FormFieldUtil.createArrayString(value);
        }
        if (typeof value === "object") {
            return FormFieldUtil.createObjectString(value);
        }
        if (asJson === true) {
            return JSON.stringify(value);
        }
        return String(value);
    },

    createArrayString: function (value) {
        var result = "[";
        for (var i = 0; i < value.length; i++) {
            if (i > 0) {
                result += ",";
            }
            var v = value[i];
            result += FormFieldUtil.getStringRepresentation(v, true);
        }
        result += "]";
        return result;
    },

    createObjectString: function (value) {
        var result = "{";
        if (value == null) {
            result += "null";
        } else {
            var keys = Object.keys(value);
            keys.sort();
            var count = 0;
            for (var i = 0; i < keys.length; i++) {
                var k = keys[i];
                var v = value[k];
                if (k.indexOf("$$")==0) {
                    continue;
                }
                if (count > 0) {
                    result += ",";
                }
                count++;
                result += FormFieldUtil.getStringRepresentation(k, true);
                result += ":";
                result += FormFieldUtil.getStringRepresentation(v, true);
            }
        }
        result += "}";
        return result;
    }
}

if (typeof String.prototype.endsWith !== 'function') {
    String.prototype.endsWith = function (suffix) {
        return this.indexOf(suffix, this.length - suffix.length) !== -1;
    };
}

/**
 * Erzeugt einen neuen String welcher aus count mal dem BasisString besteht. Nötig hier solange string.repeat nicht in jedem modernen Browser unterstützt wird. Aktuell kann es NUR FireFox.
 * http://stackoverflow.com/questions/202605/repeat-string-javascript
 */
if (!String.prototype.repeat) {
    String.prototype.repeat = function (count) {
        if (count < 1) return '';
        var result = '', pattern = this.valueOf();
        while (count > 1) {
            if (count & 1) result += pattern;
            count >>= 1, pattern += pattern;
        }
        return result + pattern;
    };
}
/**
 * Für veraltete Schrammel-browser die kein Array.find unterstützen lege ich das hier an. Aktuell können es alle außer IE.
 * https://tc39.github.io/ecma262/#sec-array.prototype.find
 */
if (!Array.prototype.find) {
  Object.defineProperty(Array.prototype, 'find', {
    value: function(predicate) {
     // 1. Let O be ? ToObject(this value).
      if (this == null) {
        throw new TypeError('"this" is null or not defined');
      }

      var o = Object(this);

      // 2. Let len be ? ToLength(? Get(O, "length")).
      var len = o.length >>> 0;

      // 3. If IsCallable(predicate) is false, throw a TypeError exception.
      if (typeof predicate !== 'function') {
        throw new TypeError('predicate must be a function');
      }

      // 4. If thisArg was supplied, let T be thisArg; else let T be undefined.
      var thisArg = arguments[1];

      // 5. Let k be 0.
      var k = 0;

      // 6. Repeat, while k < len
      while (k < len) {
        // a. Let Pk be ! ToString(k).
        // b. Let kValue be ? Get(O, Pk).
        // c. Let testResult be ToBoolean(? Call(predicate, T, « kValue, k, O »)).
        // d. If testResult is true, return kValue.
        var kValue = o[k];
        if (predicate.call(thisArg, kValue, k, o)) {
          return kValue;
        }
        // e. Increase k by 1.
        k++;
      }

      // 7. Return undefined.
      return undefined;
    },
    configurable: true,
    writable: true
  });
}

export const ValueFormatter = {

    formatFileSize: function (size) {
        var level = 0;
        while (Math.floor(size) > 512) {
            size = size / 1024;
            level++;
        }
        var locale = ($('html').attr('locale') ||  $('html').attr('lang'));
        var result = "";
        result += size.toLocaleString(locale, {
            style: 'decimal',
            minimumFractionDigits: level == 0 ? 0 : 2,
            maximumFractionDigits: level == 0 ? 0 : 2,
            useGrouping: false
        }); 
        result += " ";
        switch (level) {
            case 1:
                result += "KB";
                break;
            case 2:
                result += "MB";
                break;
            case 3:
                result += "GB";
                break;
            case 4:
                result += "TB";
                break;
            default:
                result += "Byte";
                break;
        }
        return result;
    },

    formatTime: function (time: number) {
        var h = parseInt(String(time / 60)); // Math.floor(time / 60); 
        var m = parseInt(String(time % 60));
        return (h > 0 ? h + "h " : "") + (m < 10 && h > 0 ? "0" + m : m) + "min";
    },

    momentInitialized: false,
    initializeMomentJs: function () {
        ValueFormatter.momentInitialized = true;
        var language = ($('html').attr('locale') ||  $('html').attr('lang'));
        moment.locale(language);
        moment().localeData()._calendar['sameElse'] = "lll";
    },

    formatDate: function (date, fullDetails) {
        if (!ValueFormatter.momentInitialized) {
            ValueFormatter.initializeMomentJs();
        }
        if (fullDetails === true) {
            return moment(date).format("LLLL");
        }
        return moment(date).calendar();
    }
}

export const ArrayHelper = {

    jsonDataToMap: function (input, key_key, value_key) {
        var returnValue = {};
        angular.forEach(input, function (entry) {
            var value = entry[value_key];
            returnValue[entry[key_key]] = JSON.parse(value);
        });
        return returnValue;
    },

    toMap: function (input, key_key, value_key) {
        var returnValue = {};
        angular.forEach(input, function (entry) {
            var value = entry[value_key];
            returnValue[entry[key_key]] = value;
        });
        return returnValue;
    },

    toStringValueMap: function (input, key_key, value_key) {
        var returnValue = {};
        angular.forEach(input, function (entry) {
            var value = entry[value_key];
            if (value == undefined || value == null) {
                value = "";
            } else if ($.isArray(value)) {
                var valArray = "";
                for (var i = 0; i < value.length; i++) {
                    if (i > 0) {
                        valArray += ";";
                    }
                    valArray += value[i];
                }
                value = valArray;
            } else {
                value = String(value);
            }
            returnValue[entry[key_key]] = value;
        });
        return returnValue;
    }

}

/**
*
*  Base64 encode / decode
*  http://www.webtoolkit.info/
*
**/
export const Base64Decoder = {

   // private property
   _keyStr: "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/=",

   // public method for encoding
   encode: function (input) {
       var output = "";
       var chr1, chr2, chr3, enc1, enc2, enc3, enc4;
       var i = 0;

       input = Base64Decoder._utf8_encode(input);

       while (i < input.length) {

           chr1 = input.charCodeAt(i++);
           chr2 = input.charCodeAt(i++);
           chr3 = input.charCodeAt(i++);

           enc1 = chr1 >> 2;
           enc2 = ((chr1 & 3) << 4) | (chr2 >> 4);
           enc3 = ((chr2 & 15) << 2) | (chr3 >> 6);
           enc4 = chr3 & 63;

           if (isNaN(chr2)) {
               enc3 = enc4 = 64;
           } else if (isNaN(chr3)) {
               enc4 = 64;
           }

           output = output +
               this._keyStr.charAt(enc1) + this._keyStr.charAt(enc2) +
               this._keyStr.charAt(enc3) + this._keyStr.charAt(enc4);

       }

       return output;
   },

   // public method for decoding
   decode: function (input) {
       var output = "";
       var chr1, chr2, chr3;
       var enc1, enc2, enc3, enc4;
       var i = 0;

       input = input.replace(/[^A-Za-z0-9\+\/\=]/g, "");

       while (i < input.length) {

           enc1 = this._keyStr.indexOf(input.charAt(i++));
           enc2 = this._keyStr.indexOf(input.charAt(i++));
           enc3 = this._keyStr.indexOf(input.charAt(i++));
           enc4 = this._keyStr.indexOf(input.charAt(i++));

           chr1 = (enc1 << 2) | (enc2 >> 4);
           chr2 = ((enc2 & 15) << 4) | (enc3 >> 2);
           chr3 = ((enc3 & 3) << 6) | enc4;

           output = output + String.fromCharCode(chr1);

           if (enc3 != 64) {
               output = output + String.fromCharCode(chr2);
           }
           if (enc4 != 64) {
               output = output + String.fromCharCode(chr3);
           }

       }

       output = Base64Decoder._utf8_decode(output);

       return output;

   },

   // private method for UTF-8 encoding
   _utf8_encode: function (string) {
       var utftext = "";

       for (var n = 0; n < string.length; n++) {

           var c = string.charCodeAt(n);

           if (c < 128) {
               utftext += String.fromCharCode(c);
           }
           else if ((c > 127) && (c < 2048)) {
               utftext += String.fromCharCode((c >> 6) | 192);
               utftext += String.fromCharCode((c & 63) | 128);
           }
           else {
               utftext += String.fromCharCode((c >> 12) | 224);
               utftext += String.fromCharCode(((c >> 6) & 63) | 128);
               utftext += String.fromCharCode((c & 63) | 128);
           }

       }

       return utftext;
   },

   // private method for UTF-8 decoding
   _utf8_decode: function (utftext) {
       var string = "";
       var i = 0;
       var c = 0;
       var c3 = 0;
       var c2 = 0;

       while (i < utftext.length) {

           c = utftext.charCodeAt(i);

           if (c < 128) {
               string += String.fromCharCode(c);
               i++;
           }
           else if ((c > 191) && (c < 224)) {
               c2 = utftext.charCodeAt(i + 1);
               string += String.fromCharCode(((c & 31) << 6) | (c2 & 63));
               i += 2;
           }
           else {
               c2 = utftext.charCodeAt(i + 1);
               c3 = utftext.charCodeAt(i + 2);
               string += String.fromCharCode(((c & 15) << 12) | ((c2 & 63) << 6) | (c3 & 63));
               i += 3;
           }

       }

       return string;
   }
}

/**
*
*  NumberUtil for IE constants
*
**/
export const NumberUtil = {
    MAX_SAFE_INTEGER: function () {
        if(Number.MAX_SAFE_INTEGER) {
            return Number.MAX_SAFE_INTEGER;
        }
        return 9007199254740991;
    },

    parseLocaleNumber: function(locale, nmbr) {
        var formattedNumber = 12345.6.toLocaleString(locale, {
            style: 'decimal',
            minimumFractionDigits: 1,
            maximumFractionDigits: 1,
            useGrouping: true
        });
        var tCh = formattedNumber.charAt(2);
        var dCh = formattedNumber.charAt(6);
        
        var result = (""+nmbr).split(tCh).join('').split(dCh).join('.');
        return parseFloat(result);
    }
}


